/* similar to PostgreSQL.org Documentation Style */

body {
  margin: 1em;
  padding: 1em;
  font-size: 85%;
  font-family: verdana, sans-serif;
  color: #000;
  background-color: #fff;
}

h1 {
  font-size: 1.4em;
  font-weight: bold;
  margin-top: 0em;
  margin-bottom: 0em;
}

h2 {
  font-size: 1.2em;
  margin: 1.2em 0em 1.2em 0em;
  font-weight: bold;
}

h3 {
  font-size: 1.0em;
  margin: 1.2em 0em 1.2em 0em;
  font-weight: bold;
}

h4 {
  font-size: 0.95em;
  margin: 1.2em 0em 1.2em 0em;
  font-weight: normal;
}

h5 {
  font-size: 0.9em;
  margin: 1.2em 0em 1.2em 0em;
  font-weight: normal;
}

h6 {
  font-size: 0.85em;
  margin: 1.2em 0em 1.2em 0em;
  font-weight: normal;
}

img {
  border: 0;
}

ol, ul, li {
  font-size: 1.0em;
  line-height: 1.2em;
  margin-top: 0.2em;
  margin-bottom: 0.1em; 
}

p {
  font-size: 1.0em;
  line-height: 1.2em;
  margin: 1.2em 0em 1.2em 0em;
}

li > p {
  margin-top: 0.2em;
}

pre {
  font-family: monospace;
  font-size: 1.2em;
  margin: 0em;
}

strong, b {
  font-weight: bold;
}

h1 {
  font-weight: bold; 
  color: #EC5800; 
  font-size: 1.4em;
}

h2 {
  font-weight: bold; 
  color: #666; 
  font-size: 1.2em;
}

h3 { 
  font-weight: bold; 
  color: #666; 
  font-size: 1.1em;
}

h4 { 
  color: #666; 
}

/* Text Styles */

.txtCurrentLocation {
  font-weight: bold;
}

p, ol, ul, li {
  line-height: 1.5em;
}

table.CALSTABLE {
  width: 50%;
}

table.CALSTABLE td {
  vertical-align : top;
}

/* Link Styles */

a:link                          { color:#0066A2; text-decoration: underline; }
a:visited                       { color:#004E66; text-decoration: underline; }
a:active                        { color:#0066A2; text-decoration: underline; }
a:hover                         { color:#000000; text-decoration: underline; }


pre.programlisting,
blockquote.note, 
blockquote.tip
{
    -moz-border-radius: 8px 8px 8px 8px;
    -moz-box-shadow: 3px 3px 5px #DFDFDF;
    color: black;
    margin: 1ex 0ex 1ex 1ex;
    padding: 1ex;
    border-style: solid;
    border-width: 1px;
}


/* Programlisting */
pre.programlisting {
    background-color: #F7F7F7;
    border-color: #CFCFCF;
    font-family: monospace;
    text-align: left;
}

div.note {
  margin-top: 4ex;
}

blockquote.note, 
blockquote.tip 
{
  padding-top: 0ex;
  background-color: #FDFDEE;
  border-color: #DBDBCC;
}


.literal,
.parameter,
.funcsynopsis,
.function
{
  font-size: 130%;
}

div.table table tr td {
  padding: 0.5ex;
}

div.example p b {
  font-size : 70%;
  font-style: italic;
}

div.NAVFOOTER {
  margin-top: 5ex;
}